use super::*;
use chrono::DateTime;
use chrono::Utc;
use std::collections::BTreeMap;

impl UserInteraction {
    pub(crate) fn new(wall_uid: WallUid, problem_uid: ProblemUid) -> Self {
        Self {
            wall_uid,
            problem_uid,
            is_favorite: false,
            attempted_on: BTreeMap::new(),
            is_saved: false,
        }
    }

    pub(crate) fn latest_attempt(&self) -> Option<DatedAttempt> {
        self.attempted_on.last_key_value().map(Into::into)
    }

    pub(crate) fn todays_attempt(&self) -> Option<Attempt> {
        self.latest_attempt()
            .filter(|latest_attempt| {
                let today_local_naive = chrono::Local::now().date_naive();
                let datetime_local_naive = latest_attempt.date_time.with_timezone(&chrono::Local).date_naive();
                datetime_local_naive == today_local_naive
            })
            .map(|dated_attempt| dated_attempt.attempt)
    }

    pub(crate) fn best_attempt(&self) -> Option<DatedAttempt> {
        self.attempted_on.iter().max_by_key(|(_date_time, attempt)| *attempt).map(Into::into)
    }

    pub(crate) fn attempted_on(&self) -> impl IntoIterator<Item = DatedAttempt> {
        self.attempted_on.iter().rev().map(Into::into)
    }
}

impl From<(DateTime<Utc>, Attempt)> for DatedAttempt {
    fn from(value: (DateTime<Utc>, Attempt)) -> Self {
        let (date_time, attempt) = value;
        DatedAttempt { date_time, attempt }
    }
}

impl From<&(DateTime<Utc>, Attempt)> for DatedAttempt {
    fn from(value: &(DateTime<Utc>, Attempt)) -> Self {
        let &(date_time, attempt) = value;
        DatedAttempt { date_time, attempt }
    }
}

impl From<(&DateTime<Utc>, &Attempt)> for DatedAttempt {
    fn from(value: (&DateTime<Utc>, &Attempt)) -> Self {
        let (&date_time, &attempt) = value;
        DatedAttempt { date_time, attempt }
    }
}

impl WallUid {
    pub(crate) fn create() -> Self {
        Self(uuid::Uuid::new_v4())
    }
}

impl ProblemUid {
    pub(crate) fn create() -> Self {
        Self(uuid::Uuid::new_v4())
    }
    pub(crate) fn min() -> Self {
        Self(uuid::Uuid::nil())
    }
    pub(crate) fn max() -> Self {
        Self(uuid::Uuid::max())
    }
}

impl Image {
    pub(crate) fn srcset(&self) -> String {
        self.resolutions
            .iter()
            .map(|(res, filename)| format!("/files/holds/{} {}w", filename.filename, res.width))
            .collect::<Vec<String>>()
            .join(", ")
    }
}

impl ImageUid {
    pub(crate) fn create() -> Self {
        Self(uuid::Uuid::new_v4())
    }
}

impl Attempt {
    pub(crate) fn icon(&self) -> crate::components::icons::Icon {
        use crate::components::icons::Icon;
        match self {
            Attempt::Attempt => Icon::ArrowTrendingUp,
            Attempt::Send => Icon::Trophy,
            Attempt::Flash => Icon::BoltSolid,
        }
    }
}
